<?php

/**
 * 
 * Return the list of the supported types
 * @return array
 */
function channels_client_node_get_types() {
  return array_keys(node_get_types());
}


/**
 *
 * Check if the given entity type is supported
 *
 * @param string $type The type of entity
 * @param string $bundle The bundle of the entity
 *
 * @return bool TRUE if the entity type is supported
 */
function channels_client_entity_supported($type, $bundle) {
  
  if($type == 'node') {
    $types = array_keys(node_get_types());
    if(in_array($bundle, $types)) return TRUE;
  }
  elseif($type == 'taxonomy_term') {
    $vocabularies = taxonomy_get_vocabularies();

    //watchdog('debug',  'type: ' . $type . ', bundle: ' . $bundle);

    foreach($vocabularies as $vid => $vocabulary) {
      // used to store the vocabulary machine name
      if($vocabulary->module== 'channels_' . $bundle){
        //watchdog('debug',  'type: ' . $type . ', bundle: ' . $bundle . ' supported');
        return TRUE;
      }
    }
  }
  //watchdog('debug',  'type: ' . $type . ', bundle: ' . $bundle . ' NOT supported');
  return FALSE;
}


/**
 *
 *  Test if an entity already exists
 *
 * @param int $id Entity ID
 * @param string $type Entity type (node, taxonomy_term...)
 *
 * @return bool
 *      TRUE if the entity exists
 *      FALSE if the entity doesn't exist
 */
function channels_client_entity_exists($id, $type) {

  if($type == 'node') {

    $map = channels_client_get_map($id, $type);
    if(!$map) return FALSE;
    elseif(isset($map['local_id'])) {
      $node = node_load($map['local_id']);
      if($node) return TRUE;
    }
  }
  elseif($type == 'taxonomy_term') {

    $map = channels_client_get_map($id, $type);
    if(!$map) return FALSE;
    elseif(isset($map['local_id'])) {
      $term = taxonomy_get_term($map['local_id']);
      if($term) return TRUE;
    }
  }
  return FALSE;
}


/**
 * 
 * Save a given entity in the local database
 * @param string $type The type of entity (node, user...)
 * @param int $id The ID of the entity
 * @param array $fields The entity's fields
 * 
 * @return bool Return TRUE when the entity is saved, or FALSE when an error occured.
 */
function channels_client_save_entity($type, $id, array $fields) {

  $map_exists = channels_client_get_map($id, $type);
  $map = $map_exists ? $map_exists : null;
  $client_id = 0;

  //////////////////////////////////////////////////////////////////////////////////////////////
  //                      NODE
  //////////////////////////////////////////////////////////////////////////////////////////////
  if( $type == 'node') {
    
    if(is_null($map))
      $node = new stdClass();
    else
      $node = node_load(array('nid' => $map['local_id']));

    channels_client_node_alter($fields, $node);

    module_invoke_all('channels_client_before_entity_save', $type, $node);

    node_save($node);

    if(!$node->nid){
      return FALSE;
    }

    module_invoke_all('channels_client_after_entity_save', $type, $node);

    $client_id = $node->nid;

  }
  //////////////////////////////////////////////////////////////////////////////////////////////
  //                      TAXONOMY_TERM
  //////////////////////////////////////////////////////////////////////////////////////////////
  elseif($type == 'taxonomy_term') {

    if(!isset($fields['name'])){
      watchdog('error', 'channels can\'t synchronize taxonomy term ' . $id . ': fields "name" not found');
      return FALSE;
    }

    if(!isset($fields['vocabulary:machine_name'])) {
       watchdog('error', 'channels can\'t synchronize taxonomy term ' . $id . ': vocabulary\'s machine name not found in fields list');
      return FALSE;
    }

    $vocabulary = channels_client_taxonomy_vocabulary_load_by_machine_name('channels_'.$fields['vocabulary:machine_name']['value']);

    if(!$vocabulary){
      watchdog('error', 'channels can\'t synchronize taxonomy term ' . $id . ': vocabulary "'. $bundle .'" not found' . serialize($fields['vocabulary:machine_name']));
      return FALSE;
    }

    // the term is unknown -> we create a new term
    if(is_null($map)) $term = array('vid' => $vocabulary->vid, 'weight' => 0);

    // the term already exists -> we load the term
    else $term = (array) taxonomy_get_term($map['local_id']);

    $term = module_invoke_all('channels_client_taxonomy_term_alter', $fields, $term);

    module_invoke_all('channels_client_before_entity_save', $type, $term);

    $res = taxonomy_save_term($term);
    watchdog('debug', 'term !term saved', array('!term' => $id));
    taxonomy_check_vocabulary_hierarchy((array)$vocabulary, $term);

    module_invoke_all('channels_client_after_entity_save', $type, $term);

    $client_id = $term['tid'];

  }

  if(is_null($map)) {
    if(!channels_client_save_map ($client_id, $id,$type) ) return FALSE;
  }

  if($client_id > 0) return TRUE;

  return FALSE;
}


function channels_client_save_file($file, $entity_id, $field_name) {
  
  global $user;
  
  $file = (object) $file;

  if(!isset($file->fid)) return services_error("Given file has no ID");

   // Prepare mapping
  $map_exists = channels_client_get_map($file->fid, 'file');
  $map = $map_exists ? $map_exists : null;

   // Write file data if exists
  if(isset($file->file) && $file->file != '') {

    module_invoke_all('channels_client_before_file_save', $file, $entity_id, $field_name);

    // Get the directory name for the location of the file:    
    $realpath = getcwd().'/'.file_create_path($file->filepath);

    // Build the destination folder tree if it doesn't already exists.
    if (!channels_client_file_check_directory(dirname($realpath), FILE_CREATE_DIRECTORY)) {
      return services_error("Could not create destination directory for file.");
    }

    // If filename has not been specified extract it from filepath.
    if ((!isset($file->filename)) || (empty($file->filename))) {
      $file->filename = trim(basename($file->filepath), '.');
    }
    $file->filename = file_munge_filename($file->filename, $extensions, TRUE);
    $file->filemime = file_get_mimetype($file->filename);

    // Rename potentially executable files, to help prevent exploits.
    if (preg_match('/\.(php|pl|py|cgi|asp|js)$/i', $file->filename) && (substr($file->filename, -4) != '.txt')) {
      $file->filemime = 'text/plain';
      $file->filepath .= '.txt';
      $file->filename .= '.txt';
    }

    // Build the real path to the file and replace it if necessary    
    $destination = file_destination($realpath, FILE_EXISTS_REPLACE);

    // If the destination is not provided, or is not writable, error our
    if (empty($file->filepath) || file_check_path($destination) === FALSE) {
      return services_error("Destination directory does not exist or is not writeable.");
    }

    if(!channels_client_save_file_data($file->fid, $file->file, $realpath))
      return services_error("Could not write file to destination");

    module_invoke_all('channels_client_after_file_save', $file, $entity_id, $field_name);

  }

  $update = array();
  if(is_null($map)) {
      $fid = $file->fid;   // backup original fid
      unset($file->fid);  // destroy original fid (as a new file)
  }
  else{
    $update = 'fid';
    $file->fid = $map['local_id'];
  }

  // If we made it this far it's safe to record this file in the database.
  if(drupal_write_record('files', $file, $update)) {

    // there's no map -> we need to create one for the current file
    if(is_null($map)) channels_client_save_map ($file->fid, $fid, 'file');
  }

  // Return the fid
  return $file->fid;
}


function channels_client_save_file_data($fid, $base64_data, $path) {

  // Write the file
  if (!file_save_data(base64_decode($base64_data), $path)) {
    watchdog('error', 'unable to save file data for file !fid', array('!fid' => $fid));
    return FALSE;
  }

  return TRUE;
}



function channels_client_file_check_directory(&$directory, $mode = 0) {

 $directory = rtrim($directory, '/\\');

  // Check if directory exists.
  if (!is_dir($directory)) {
    if (($mode & FILE_CREATE_DIRECTORY) && @mkdir($directory, 0777, TRUE)) {
      @chmod($directory, 0775); // Necessary for non-webserver users.
    }
    else {
      return FALSE;
    }
  }

  // Check to see if the directory is writable.
  if (!is_writable($directory)) {
    if (!($mode & FILE_MODIFY_PERMISSIONS) && @chmod($directory, 0775)) {
      watchdog('file system', 'The directory %directory is not writable, because it does not have the correct permissions set.', array('%directory' => $directory), WATCHDOG_ERROR);
      return FALSE;
    }
  }
  
  return TRUE;
}


/**
 *
 * Delete a given entity in the local database
 * @param string $type The type of entity (node, user...)
 * @param int $id The ID of the entity
 * @param array $fields The entity's fields
 *
 * @return bool Return TRUE when the entity is saved, or FALSE when an error occured.
 */
function channels_client_delete_entity($type, $id) {

  $map_exists = channels_client_get_map($id, $type);
  $map = $map_exists ? $map_exists : null;

  $skip_delete = module_invoke_all('channels_client_before_entity_delete', $type, (int) $map['local_id']);

  //////////////////////////////////////////////////////////////////////////////////////////////
  //                      NODE
  //////////////////////////////////////////////////////////////////////////////////////////////
  if( $type == 'node') {

    if(!is_null($map)) {
      
      if(!$skip_delete) node_delete($map['local_id']);
      module_invoke_all('channels_client_after_entity_delete', $type, (int) $map['local_id']);
      return TRUE;
    }
    

  }
  //////////////////////////////////////////////////////////////////////////////////////////////
  //                      TAXONOMY_TERM
  //////////////////////////////////////////////////////////////////////////////////////////////
  elseif($type == 'taxonomy_term') {

    if(!is_null($map)) {

      if(!$skip_delete  && taxonomy_del_term ($map['local_id']) == SAVED_DELETED ) {
        module_invoke_all('channels_client_after_entity_delete', $type, (int) $map['local_id']);
        return TRUE;
      }
    }
  }
  return FALSE;

}


/**
 * 
 * Alter a node before the client save it (convert a drupal 7 node to a drupal 6 one)
 *
 * @param array $fields List of the node's fields received by the service
 * @param object $node Object representation of a node
 */
function channels_client_node_alter($fields, $node) {

  $excluded_fields = array('is_new', 'process_api_language', 'channels_options');

  /******************** Filter format **********************************************************/
  $formats = filter_formats();

  $body_fields = array();
  foreach ($fields as $field_key => $field) {
    if (preg_match('/^body\:/', $field_key)) {
     $body_fields[$field_key] = $field['value'];
     unset($node->{$field_key});
    }
    elseif($field_key != 'nid' && $field_key != 'vid' && !in_array($field_key, $excluded_fields)) {
      $node->{$field_key} = $field['value'];
    }
  }

 foreach ($body_fields as $key => $value) {
    $ids = preg_split('/body:/', $key);
    if (isset($ids[1]) ) {
      switch ($ids[1]) {
        case 'value' :  
          $node->body = $value;
          break;
          
        case 'summary' :   
          $node->teaser = $value;
          $node->body = $value . '<!--break-->' . $node->body;
          break;
        case 'format' :  
          if ($value == 'plain_text') $node->format = 0;
          elseif ($value == 'filtered_html') $node->format = 1;
          elseif ($value == 'full_html') $node->format = 2;
          break;
      }
    }
  }

  module_invoke_all('channels_client_node_alter', $fields, $node);

  return $node;
}


/**
 *
 * Alter a taxonomy term before the client save it (convert a drupal 7 term to a drupal 6 one)
 *
 * @param array $fields List of the term's fields received by the service
 * @param array $term Array representation of a term
 */
function channels_client_channels_client_taxonomy_term_alter($fields, $term) {

  // term properties update
  $term['name'] = $fields['name']['value'];
  if(isset($fields['description'])) $term['description'] = $fields['description']['value'];
  if(isset($fields['weight'])) $term['weight'] = $fields['weight']['value'];

  // parent terms
  $term['parent'] = array();
  if(isset($fields['parent:tid']) && is_array($fields['parent:tid']['value'])) {
    foreach($fields['parent:tid']['value'] as $parent_tid) {
      $map_parent = channels_client_get_map($parent_tid, 'taxonomy_term');
      if(is_null($map_parent))
        watchdog('error', 'no parent map found for parent !parent of the term !term', array('!term' => $id, '!parent' => $parent_tid));
      else
        $term['parent'][] = $map_parent['local_id'];
    }
  }
  return $term;
}

/**
 *
 *  Return the identifiers' map for a given entity
 *
 * @param int $external_id The external id of an entity
 * @param string $entity_type The type of the entity
 * @return A database query result array, or FALSE if the query was not executed correctly.
 */
function channels_client_get_map($external_id, $entity_type) {
  if( !is_numeric($external_id) ) return false;
  $res = db_query('select * from {channels_client_map} where external_id = \'%s\' and type = \'%s\'', array($external_id, $entity_type));
  return db_fetch_array($res);
}

/**
 *
 *  Save the identifiers' map for a given entity
 *
 * @param string $local_id Local identifier of the entity
 * @param int $external_id External identifier of the entity
 * @param string $type The type of the entity (node | taxonomy_term)
 */
function channels_client_save_map($local_id, $external_id, $type) {
  module_invoke_all('channels_client_presave_map', $local_id, $external_id, $type);
  $res = db_query('insert into {channels_client_map}(local_id, external_id, type) values (\'%s\', %d, \'%s\')', $local_id, $external_id, $type);
  module_invoke_all('channels_client_save_map', $local_id, $external_id, $type);
  if(!$res) return FALSE;
  return TRUE;
}

/**
 * Remove the identifiers' map for the given local entity
 *
 * @param int $local_id Local node id
 * @param string $type Type of entity (node, taxonomy_term...)
 * @return bool
 *  Return TRUE if the map is successfully deleted
 *  Return FALSE if an error occured
 */
function channels_client_delete_map($local_id, $type) {
  $res = db_query('delete from {channels_client_map} where local_id = ' . $local_id .' and type = \'' . $type . '\'' );
  module_invoke_all('channels_client_delete_map', $local_id, $type);
  if(!$res) return FALSE;
  return TRUE;
}

/**
 *
 * Implements hook_nodeapi
 * 
 */
function channels_client_nodeapi(&$node, $op, $a3 = NULL, $a4 = NULL) {
  if($op == 'delete' && !channels_client_delete_map($node->nid, 'node') ) {
    watchdog('error', 'unable to delete the map for the entity id: ' . $node->nid);
  }
}

/**
 * Targeted form_alter for vocabulary edit pages.
 */
function channels_client_form_taxonomy_form_vocabulary_alter(&$form, &$form_state) {
  
  if(module_exists('features')) return;

  $elem = array(
    '#type' => 'textfield',
    '#title' => t('Machine name'),
    '#description' => t('A system named for this vocabulary, may not be changed once it has been set.'),
  );

  if (!isset($form['module']) || $form['module']['#value'] == 'taxonomy') {
    // This vocabulary is new or wasn't assigned a machine name.
    $elem['#element_validate'] = array('channels_client_taxonomy_machine_name_validate');
    unset($form['module']);
  }
  elseif (strpos($form['module']['#value'], 'channels_') === 0) {
    // Simply display the existing machine name if we have one.
    $elem['#disabled'] = true;
    $elem['#value'] = substr($form['module']['#value'], 9);
  }
  else {
    // Some other module is responsible here, do nothing.
    return;
  }

  // Position the machine name right after the vocab name without interfering
  // with any explicit weights.
  $new = array();
  foreach ($form['identification'] as $k => $v) {
    $new[$k] = $v;
    if ($k == 'name') {
      $new['module'] = $elem;
    }
  }
  $form['identification'] = $new;

}

// Validation handler for the machine name field.
function channels_client_taxonomy_machine_name_validate($element, &$form_state) {

  if(module_exists('features')) return;

  // Validate format - lowecase + underscore only...
  if (!empty($element['#value']) && !preg_match('!^[a-z0-9_]+$!', $element['#value'])) {
    form_error($element, t('The machine name can only consist of lowercase letters, underscores, and numbers.'));
  }

  // If provided check machine name for conflicts and prepend with 'features'.
  if (!empty($element['#value'])) {
    $name = 'channels_' . $element['#value'];
    $vid = db_result(db_query("SELECT vid FROM {vocabulary} WHERE module = '%s'", $name));
    if ($vid) {
      form_error($element, t('This machine name is already taken. Please enter a unique machine name or this taxonomy.'));
    }
    else {
      $form_state['values']['module'] = $name;
    }
  }
  // Otherwise, default to the default module name 'taxonomy'.
  else {
    $form_state['values']['module'] = 'taxonomy';
  }
}


/**
 * Return the vocabulary object matching a vocabulary machine name.
 *
 * @param $machine_name
 *   The vocabulary's machine_name
 * @param $reset
 *   Whether to reset the internal taxonomy_vocabulary_load cache.
 *
 * @return
 *   The vocabulary object with all of its metadata, if exists, FALSE otherwise.
 *   Results are statically cached.
 */
function channels_client_taxonomy_vocabulary_load_by_machine_name($machine_name, $reset = FALSE) {
  
  static $vocabularies = array();

  if ($reset) {
    $vocabularies = array();
  }

  if (!isset($vocabularies[$vid])) {
    // Initialize so if this vocabulary does not exist, we have
    // that cached, and we will not try to load this later.
    $vocabularies[$vid] = FALSE;
    // Try to load the data and fill up the object.
    $result = db_query('SELECT v.*, n.type FROM {vocabulary} v LEFT JOIN {vocabulary_node_types} n ON v.vid = n.vid WHERE v.module = \'%s\'', $machine_name);

    $node_types = array();
    while ($voc = db_fetch_object($result)) {
      if (!empty($voc->type)) {
        $node_types[$voc->type] = $voc->type;
      }
      unset($voc->type);
      $voc->nodes = $node_types;
      $vocabularies[$vid] = $voc;
    }
  }

  // Return FALSE if this vocabulary does not exist.
  return !empty($vocabularies[$vid]) ? $vocabularies[$vid] : FALSE;
}

/**
 * Implements hook_taxonomy()
 */
function channels_client_taxonomy($op, $type, $array) {
  if($type == 'term') {
    switch($op) {
      case 'delete':
        if(!channels_client_delete_map($array['tid'], 'taxonomy_term')) watchdog('error', 'unable to delete the map for the entity id: ' . $node->nid);
        break;
    }
  }
}